@@ -1,7 +1,7 @@
 ;;;
 ;;; gauce.cgen.stub - stub forms parsing and code generation
 ;;;
-;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -121,7 +121,7 @@
 ;;      <Qual> (qualifier) is a list to adds auxiliary information to the
 ;;      SUBR.  Currently the following <quals> are officially supported.
 ;;
-;;        (setter <setter-name>) : specfy setter.  <setter-name> should
+;;        (setter <setter-name>) : specify setter.  <setter-name> should
 ;;             be a cproc name defined in the same stub file
 ;;        (setter (args ...) body ...) : specify setter anonymously.
 ;;
@@ -131,7 +131,7 @@
 ;;             Gauche error condition.
 ;;
 ;;        (inliner <insn-name>) : only used in Gauche core procedures
-;;             that can be inlined into an VM insturuction.
+;;             that can be inlined into an VM instruction.
 ;;
 ;;      <Body> is a cise expression.  Inside the expression, a cise macro
 ;;      (result expr ...) can be used to assign the value(s) to return
@@ -243,7 +243,7 @@
                       :name (path-sans-extension (sys-basename stubfile))
                       :c-name-prefix (string-append prefix "_")
                       :preamble '("/* Generated by genstub.  Do not edit. */")
-                      :init-prologue (format "void Scm_Init_~a(ScmModule *mod)\
+                      :init-prologue (format "void Scm_Init_~a(ScmModule *mod SCM_UNUSED)\
                                               {\n" prefix))])
       (with-tmodule-recording
        <tmodule>
@@ -292,6 +292,14 @@
   `(make <form-parser> :name ',name :args ',args
          :handler (^ (&whole ,@args) ,@body)))
 
+(define-syntax export-toplevel-cise-form
+  (syntax-rules ()
+    [(_ name)
+     (define-form-parser name args
+       (parameterize ([cise-ambient (cise-ambient-copy (cise-ambient) '())])
+         (cgen-body (cise-render-to-string `(name ,@args) 'toplevel))
+         (cgen-decl (cise-ambient-decl-strings (cise-ambient)))))]))
+
 (define-method invoke ((self <form-parser>) form)
   (define (badform)
     (errorf <cgen-stub-error> "stub: malformed ~a: ~s" (~ self'name) form))
@@ -326,8 +334,18 @@
 (define-form-parser define-cise-expr args
   (eval `(define-cise-expr ,@args) (current-module)))
 
-(define-form-parser define-cfn args
-  (cgen-body (cise-render-to-string `(define-cfn ,@args) 'toplevel)))
+;; toplevel CiSE forms
+(export-toplevel-cise-form declare-cfn)
+(export-toplevel-cise-form declare-cvar)
+(export-toplevel-cise-form define-cfn)
+(export-toplevel-cise-form define-ctype)
+(export-toplevel-cise-form define-cvar)
+
+; CiSE forms .if, .cond, .when and .unless are not exported because
+; there are (or will be) stub replacements of the same name.
+(export-toplevel-cise-form .define)
+(export-toplevel-cise-form .include)
+(export-toplevel-cise-form .undef)
 
 ;; extra check of valid clauses
 (define (check-clauses directive name clauses valid-keys)
@@ -426,22 +444,29 @@
    ;; - <string>: C variable name for unboxed value
    (scm-name)
    ;; - <string>: C variable name to hold boxed ScmObj value
-   (count    :init-keyword :count)
-   ;; - <integer>: This arg is count-th in the procedure
    (type     :init-keyword :type)
    ;; - <cgen-type>: Stub type of this arg
    (default  :init-keyword :default :init-value #f)
    ;; - #f or <cgen-literal> : default value for optional/keyword arg
    ))
 
+;; required, optional and optarray arg uses "count" slot, that indicates
+;; the arg is count-th argument in the procedure.
 ;; The opt-count is an integer, indicating the position of this optional
 ;; arg counted from the beginning of the optional arg
-(define-class <required-arg> (<arg>) ())
-(define-class <optional-arg> (<arg>) ((opt-count :init-keyword :opt-count)))
-(define-class <keyword-arg>  (<arg>) (keyword))
-(define-class <rest-arg>     (<arg>) ())
-(define-class <optarray-arg> (<arg>) ((count-var :init-keyword :count-var)
-                                      (max-count :init-keyword :max-count)))
+(define-class <required-arg> (<arg>)
+  ((count     :init-keyword :count)))
+(define-class <optional-arg> (<arg>)
+  ((count     :init-keyword :count)
+   (opt-count :init-keyword :opt-count)))
+(define-class <keyword-arg>  (<arg>)
+  (keyword))
+(define-class <rest-arg>     (<arg>)
+  ())
+(define-class <optarray-arg> (<arg>)
+  ((count     :init-keyword :count)
+   (count-var :init-keyword :count-var)
+   (max-count :init-keyword :max-count)))
 
 (define-method write-object ((self <arg>) out)
   (format out "#<~a ~a>" (class-of self) (~ self'name)))
@@ -533,8 +558,8 @@
    (num-reqargs       :initform 0   :init-keyword :num-reqargs)
       ;; # of required arguments.
    (num-optargs       :initform 0   :init-keyword :num-optargs)
-      ;; # of optional arguments.  including keyword-args.  rest-arg
-      ;; is counted as 1 if the procedure takes it.
+      ;; # of optional arguments.  keyword args and rest-args
+      ;; are counted as 1 if the procedure takes it.
    (have-rest-arg?    :initform #f  :init-keyword :have-rest-arg?)
       ;; true if the procedure takes a rest-arg.
    (allow-other-keys? :initform '() :init-keyword :allow-other-keys?)
@@ -557,7 +582,11 @@
       ;; list of strings for code that must be emitted before the
       ;; procedure definition
    (source-info       :initform #f :init-keyword :source-info)
-      ;; (file . line-no), if known.
+      ;; (file::<string> line-no::<integer>), if known.
+      ;; Will be saved in pair attributes of info.
+   (bind-info         :initform #f :init-keyword :bind-info)
+      ;; (module-name::<symbol> proc-name::<symbol>), if known.
+      ;; Will be saved in pair attributes of info.
    (info              :initform #f)
       ;; cgen-literal of ScmObj to be set in the 'info' slot of ScmProcedure.
    ))
@@ -572,6 +601,28 @@
 ;; This parameter holds <procstub> during processing the body.
 (define current-procstub (make-parameter #f))
 
+;; Contains a list of return type objects the current cise expression
+;; supposed to generate.  This may differ from (~ (current-procstub)'rettype),
+;; for the cise expression may call a C routine that generates multiple
+;; values directly on VM.  In that case, this parameter contains
+;; just (#<type top>).   () means no return value.  #f means no information.
+(define cise-return-types (make-parameter #f))
+
+;; Given rettype : #f | symbol | (symbol ...),
+;; bind cise-return-types with (Type ...)
+(define-syntax with-return-types
+  (syntax-rules ()
+    [(_ rettype body ...)
+     (parameterize ([cise-return-types
+                     (match rettype
+                       [#f `(,*scm-type*)]
+                       ['<void> '()]
+                       [(? symbol?) `(,(name->type rettype))]
+                       [(? list?) (map name->type rettype)]
+                       [_ (error <cgen-stub-error>
+                                 "invalid cproc return type:" rettype)])])
+       body ...)]))
+
 ;; API
 ;; We use customized cise context while generating stubs
 (define (cgen-stub-cise-ambient :optional (orig-ambient (cise-ambient)))
@@ -579,18 +630,23 @@
     ;; Override "return"
     ($ cise-register-macro! 'return
        (^[form env]
+         (when (and (cise-return-types)
+                    (not (= (length (cdr form)) 
+                            (length (cise-return-types)))))
+           (errorf <cgen-stub-error> 
+                   "return macro got wrong number of return values (expecting ~s): ~s" 
+                   (length (cise-return-types)) form))
          (match form
            [(_)         `(goto SCM_STUB_RETURN)]
            [(_ e0)       `(begin (set! SCM_RESULT ,e0)
                                  (goto SCM_STUB_RETURN))]
-           [(_ e0 e1)    `(begin (set! SCM_RESULT0 ,e0)
-                                 (set! SCM_RESULT1 ,e1)
-                                 (goto SCM_STUB_RETURN))]
-           [(_ e0 e1 e2) `(begin (set! SCM_RESULT0 ,e0)
-                                 (set! SCM_RESULT1 ,e1)
-                                 (set! SCM_RESULT2 ,e2)
-                                 (goto SCM_STUB_RETURN))]
-           [_ (error "Too many values to return")]))
+           [(_ xs ...) `(begin 
+                          (set!
+                           ,@(concatenate
+                              (map-with-index
+                               (^[i x] `(,(string->symbol #"SCM_RESULT~i") ,x))
+                               xs)))
+                          (goto SCM_STUB_RETURN))]))
        ctx)))
 
 (define-syntax with-cise-ambient
@@ -602,6 +658,7 @@
        (push! (~ procstub'forward-decls)
               (cise-ambient-decl-strings (cise-ambient))))]))
 
+
 ;;-----------------------------------------------------------------
 ;; (define-cproc scheme-name (argspec) body)
 ;;
@@ -650,6 +707,10 @@
                       :source-info (or (debug-source-info &whole)
                                        (debug-source-info (cdr &whole))
                                        (debug-source-info (caddr &whole)))
+                      :bind-info (and-let* ([m (current-tmodule)]
+                                            [mod-name (~ m'name)]
+                                            [ (symbol? mod-name) ])
+                                   (list mod-name scheme-name))
                       :keyword-args keyargs
                       :num-reqargs nreqs
                       :num-optargs nopts
@@ -704,8 +765,7 @@
     (errorf "~a form is not in define-cproc" who)))
 
 (define (%expand-funcall-apply applyproc tmpvar args)
-  (define (suffix n) ; NB: symbol-append is not in 0.9.4 yet
-    (string->symbol #"~|applyproc|~|n|"))
+  (define (suffix n) (string->symbol #"~|applyproc|~|n|"))
   (match args
     [()           `(,(suffix 0) ,tmpvar)]
     [(a)          `(,(suffix 1) ,tmpvar ,a)]
@@ -734,7 +794,8 @@
    (let1 tmpvar (%insert-funcall-decls '.funcall/cps scheme-proc-name #f)
      (%expand-funcall-apply 'Scm_VMApply tmpvar arg))])
 
-;; create arg object.  used in cproc and cmethod
+;; create arg object.  used in cproc and cmethod 
+;; NB: count arg is ignored if class is <keyword-arg> or <rest-arg>
 (define (make-arg class argname count . rest)
   (define (grok-argname argname)
     (let1 namestr (symbol->string argname)
@@ -766,18 +827,18 @@
     (match specs
       [()                  (values (reverse args) '() nreqs 0 #f #f)]
       [(':optional . specs) (optional specs args nreqs 0)]
-      [(':rest . specs)     (rest specs args '() nreqs 0 #f)]
-      [(':key . specs)      (keyword specs args '() nreqs 0)]
+      [(':rest . specs)     (rest specs args '() nreqs 1 #f)]
+      [(':key . specs)      (keyword specs args '() nreqs 1)]
       [(':allow-other-kyes . specs)
        (errorf <cgen-stub-error>
                "misplaced :allow-other-key parameter: ~s in ~a" argspecs name)]
-      [(':optarray (var cnt max) . specs)
+      [(':optarray (var cnt maxcnt) . specs)
        (let1 args (cons (make-arg <optarray-arg> var nreqs
-                                  :count-var cnt :max-count max)
+                                  :count-var cnt :max-count maxcnt)
                         args)
          (match specs
-           [() (values (reverse args) '() nreqs max #f #f)]
-           [(':rest . specs) (rest specs args '() nreqs max #f)]
+           [() (values (reverse args) '() nreqs maxcnt #f #f)]
+           [(':rest . specs) (rest specs args '() nreqs (+ maxcnt 1) #f)]
            [_ (badarg specs)]))]
       [([? symbol? sym] . specs)
        (required specs
@@ -796,7 +857,7 @@
       [(':optarray . specs)
        (error <cgen-stub-error>
               ":optarray and :optional can't be used together in "name)]
-      [(':rest . specs)     (rest specs args '() nreqs nopts #f)]
+      [(':rest . specs)     (rest specs args '() nreqs (+ nopts 1) #f)]
       [(':allow-other-keys . specs)
        (error <cgen-stub-error>
               "misplaced :allow-other-keys parameter in "name)]
@@ -837,24 +898,24 @@
       [(':rest . specs)     (rest specs args keyargs nreqs nopts #f)]
       [([? symbol? sym] . specs)
        (keyword specs args
-                (cons (make-arg <keyword-arg> sym (+ nreqs nopts))
+                (cons (make-arg <keyword-arg> sym 0)
                       keyargs)
-                nreqs (+ nopts 1))]
+                nreqs nopts)]
       [(([? symbol? sym] default) . specs)
        (keyword specs args
-                (cons (make-arg <keyword-arg> sym (+ nreqs nopts)
+                (cons (make-arg <keyword-arg> sym 0
                                 :default (make-literal default))
                       keyargs)
-                nreqs (+ nopts 1))]
+                nreqs nopts)]
       [_ (badarg (car specs))]))
 
   (define (rest specs args keyargs nreqs nopts other-keys?)
     (match specs
       [() (values (reverse args) (reverse keyargs) nreqs nopts #t other-keys?)]
       [((? symbol? sym))
-       (values (reverse (cons (make-arg <rest-arg> sym (+ nreqs nopts)) args))
+       (values (reverse (cons (make-arg <rest-arg> sym 0) args))
                (reverse keyargs)
-               nreqs (+ nopts 1) #t other-keys?)]
+               nreqs nopts #t other-keys?)]
       [_ (badarg (car specs))]))
 
   (required (xlate-old-lambda-keywords argspecs) '() 0)
@@ -1067,20 +1128,16 @@
                       [else (cgen-return-stmt
                              #`"Scm_Values(Scm_List(,results,, NULL))")]))
         )))
-  ($ with-cise-ambient cproc
-     (match rettype
-       [#f             ; the default case; we assume it is <top>.
-        (typed-result *scm-type* body)]
-       ['<void>        ; no results
+  ($ with-return-types rettype
+     $ with-cise-ambient cproc
+     (match (cise-return-types)
+       [()
         (for-each expand-stmt body)
         (push-stmt! cproc "goto SCM_STUB_RETURN;") ; avoid 'label not used' error
         (push-stmt! cproc "SCM_STUB_RETURN:") ; label
         (push-stmt! cproc "SCM_RETURN(SCM_UNDEFINED);")]
-       [[? symbol? t]  ; single result
-        (typed-result (name->type t) body)]
-       [[? list? ts]   ; multiple values
-        (typed-results (map name->type ts) body)]
-       [_ (error <cgen-stub-error> "invalid cproc return type:" rettype)])))
+       [(type) (typed-result type body)]
+       [[types ...] (typed-results types body)])))
 
 (define-method assign-proc-info! ((proc <cproc>))
   (define (arginfo arg) (~ arg'name))
@@ -1097,9 +1154,11 @@
                         [(pair? aarg) @ `(:optarray ,(arginfo (car aarg)))]))])
     (set! (~ proc'info)
           (make-literal
-           (if (~ proc'source-info)
+           (if (or (~ proc'source-info) (~ proc'bind-info))
              ($ make-serializable-extended-pair (~ proc'scheme-name) all-args
-                `((source-info . ,(~ proc'source-info))))
+                (cond-list
+                 [(~ proc'source-info) => (cut cons 'source-info <>)]
+                 [(~ proc'bind-info)   => (cut cons 'bind-info <>)]))
              (cons (~ proc'scheme-name) all-args))))))
 
 ;;;
@@ -1112,7 +1171,7 @@
 (define-method cgen-emit-decl ((cproc <cproc>))
   (next-method)
   ;; We emit forward definitions of C function names and stub record definitions
-  ;; first, so that the funciton body can refer to the stub record.
+  ;; first, so that the function body can refer to the stub record.
   (p "static ScmObj "(~ cproc'c-name)"(ScmObj*, int, void*);")
   (let1 flags (~ cproc'flags)
     (format #t "static SCM_DEFINE_SUBR~a(" (if (null? flags) "" "X"))
@@ -1139,7 +1198,7 @@
   (p))
 
 (define-method cgen-emit-body ((cproc <cproc>))
-  (p "static ScmObj "(~ cproc'c-name)"(ScmObj *SCM_FP, int SCM_ARGCNT, void *data_)")
+  (p "static ScmObj "(~ cproc'c-name)"(ScmObj *SCM_FP SCM_UNUSED, int SCM_ARGCNT SCM_UNUSED, void *data_ SCM_UNUSED)")
   (p "{")
   ;; argument decl
   (for-each emit-arg-decl (~ cproc'args))
@@ -1212,7 +1271,8 @@
     [(? symbol? x)
      (or (and-let* ([setter (find (^z (eq? (~ z'scheme-name) x))
                                   (get-stubs <stub>))])
-           (emit (c-stub-name setter)))
+           (emit (c-stub-name setter))
+           #t)
          (errorf <cgen-stub-error>
                  "unknown setter name '~a' is used in the definition of '~a'"
                  x (~ cproc'scheme-name)))]
@@ -1408,7 +1468,7 @@
         ))))
 
 (define-method cgen-emit-body ((method <cmethod>))
-  (f "static ScmObj ~a(ScmNextMethod *nm_, ScmObj *SCM_FP, int SCM_ARGCNT, void *d_)"
+  (f "static ScmObj ~a(ScmNextMethod *nm_ SCM_UNUSED, ScmObj *SCM_FP, int SCM_ARGCNT SCM_UNUSED, void *d_ SCM_UNUSED)"
      (~ method'c-name))
   (p "{")
   (for-each emit-arg-decl (~ method'args))
@@ -1508,7 +1568,7 @@
 ;; qualifiers := :base | :built-in | :private
 
 ;; 'qualifiers' modifies the generated code slightly.  :base and :built-in
-;; is exclusive.  :base generates a base class definition (inheritable
+;; are exclusive.  :base generates a base class definition (inheritable
 ;; from Scheme code), while :built-in generates a built-in class definition
 ;; (not inheritable from Scheme code).  If neither one appears, built-in
 ;; class is generated.  :private is an optional qualifier, and when it is
@@ -1621,7 +1681,7 @@
     (p "}")
     (p ""))
   (unless ((any-pred not c-literal?) (~ self'printer))
-    (p "static void "(c-printer-name self)"(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)")
+    (p "static void "(c-printer-name self)"(ScmObj obj, ScmPort *port, ScmWriteContext *ctx SCM_UNUSED)")
     (p "{")
     (p (c-code (~ self'printer)))
     (p "}")
@@ -1773,7 +1833,7 @@
 ;;              subclass of <foreign-pointer> to represent this C-ptr type.
 ;;  c-type    : Name of the actual C type we wrap.
 ;;  c-name    : A C variable name (must have type ScmClass *).  If :private
-;;              qualifier is given, static definition of the varaible is
+;;              qualifier is given, the static definition of the variable is
 ;;              generated; otherwise, the definition & declaration must
 ;;              be provided elsewhere.  In initialization code, an instance
 ;;              of a class (the same one bound to scm-name in the Scheme
@@ -1879,6 +1939,7 @@
 ;; Main parsers
 ;;
 
+;; deprecated
 (define-form-parser if (test then . maybe-else)
   (cgen-with-cpp-condition test
     (cgen-stub-parse-form then))
@@ -1886,13 +1947,29 @@
     (cgen-with-cpp-condition `(not ,test)
       (cgen-stub-parse-form (car maybe-else)))))
 
+(define-form-parser .if (test then . maybe-else)
+  (cgen-with-cpp-condition test
+    (cgen-stub-parse-form then))
+  (unless (null? maybe-else)
+    (cgen-with-cpp-condition `(not ,test)
+      (cgen-stub-parse-form (car maybe-else)))))
+
 (define-form-parser begin forms
   (for-each cgen-stub-parse-form forms))
 
+;; deprecated
 (define-form-parser when (test . forms)
   (cgen-with-cpp-condition test
     (for-each cgen-stub-parse-form forms)))
 
+(define-form-parser .when (test . forms)
+  (cgen-with-cpp-condition test
+    (for-each cgen-stub-parse-form forms)))
+
+(define-form-parser .unless (test . forms)
+  (cgen-with-cpp-condition `(not ,test)
+    (for-each cgen-stub-parse-form forms)))
+
 (define-form-parser include (file)
   (unless (file-exists? file)
     ;; TODO: search path
